@@ -5,7 +5,7 @@
 //!
 //! 1. The source medium of the deserialization, e.g. whether the data is serialized from a `[u8]` slice, or some other container
 //! 2. The format of the deserialization, such as if the original data is encoded in a COBS format, contains a CRC32 checksum
-//!      appended to the message, etc.
+//!    appended to the message, etc.
 //!
 //! Flavors are implemented using the [`Flavor`] trait, which acts as a "middleware" for retrieving the bytes before they
 //! are passed to `serde` for deserialization
@@ -116,9 +116,31 @@ pub trait Flavor<'de>: 'de {
         None
     }
 
-    /// Attempt to take the next `ct` bytes from the serialized message
+    /// Attempt to take the next `ct` bytes from the serialized message.
+    ///
+    /// This variant borrows the data from the input for zero-copy deserialization. If zero-copy
+    /// deserialization is not necessary, prefer to use `try_take_n_temp` instead.
     fn try_take_n(&mut self, ct: usize) -> Result<&'de [u8]>;
 
+    /// Attempt to take the next `ct` bytes from the serialized message.
+    ///
+    /// This variant does not guarantee that the returned value is borrowed from the input, so it
+    /// cannot be used for zero-copy deserialization, but it also avoids needing to potentially
+    /// allocate a data in a temporary buffer.
+    ///
+    /// This variant should be used instead of `try_take_n`
+    /// if zero-copy deserialization is not necessary.
+    ///
+    /// It is only necessary to implement this method if the flavor requires storing data in a
+    /// temporary buffer in order to implement the borrow semantics, e.g. the `std::io::Read`
+    /// flavor.
+    fn try_take_n_temp<'a>(&'a mut self, ct: usize) -> Result<&'a [u8]>
+    where
+        'de: 'a,
+    {
+        self.try_take_n(ct)
+    }
+
     /// Complete the deserialization process.
     ///
     /// This is typically called separately, after the `serde` deserialization
@@ -138,9 +160,10 @@ pub struct Slice<'de> {
 impl<'de> Slice<'de> {
     /// Create a new [Slice] from the given buffer
     pub fn new(sli: &'de [u8]) -> Self {
+        let range = sli.as_ptr_range();
         Self {
-            cursor: sli.as_ptr(),
-            end: unsafe { sli.as_ptr().add(sli.len()) },
+            cursor: range.start,
+            end: range.end,
             _pl: PhantomData,
         }
     }
@@ -155,6 +178,8 @@ impl<'de> Flavor<'de> for Slice<'de> {
         if self.cursor == self.end {
             Err(Error::DeserializeUnexpectedEnd)
         } else {
+            // SAFETY: `self.cursor` is in-bounds and won't be incremented past `self.end` as we
+            // have checked above.
             unsafe {
                 let res = Ok(*self.cursor);
                 self.cursor = self.cursor.add(1);
@@ -174,6 +199,8 @@ impl<'de> Flavor<'de> for Slice<'de> {
         if remain < ct {
             Err(Error::DeserializeUnexpectedEnd)
         } else {
+            // SAFETY: `self.cursor` is valid for `ct` elements and won't be incremented past `self.end` as we
+            // have checked above.
             unsafe {
                 let sli = core::slice::from_raw_parts(self.cursor, ct);
                 self.cursor = self.cursor.add(ct);
@@ -185,11 +212,12 @@ impl<'de> Flavor<'de> for Slice<'de> {
     /// Return the remaining (unused) bytes in the Deserializer
     fn finalize(self) -> Result<&'de [u8]> {
         let remain = (self.end as usize) - (self.cursor as usize);
+        // SAFETY: `self.cursor` is valid for `remain` elements
         unsafe { Ok(core::slice::from_raw_parts(self.cursor, remain)) }
     }
 }
 
-/// Support for [std::io] or `embedded-io` traits
+/// Support for [`std::io`] or `embedded-io` traits
 #[cfg(any(
     feature = "embedded-io-04",
     feature = "embedded-io-06",
@@ -207,24 +235,22 @@ pub mod io {
 
     impl<'de> SlidingBuffer<'de> {
         pub fn new(sli: &'de mut [u8]) -> Self {
+            let range = sli.as_mut_ptr_range();
             Self {
-                cursor: sli.as_mut_ptr(),
-                end: unsafe { sli.as_ptr().add(sli.len()) },
+                cursor: range.start,
+                end: range.end,
                 _pl: PhantomData,
             }
         }
 
-        #[inline]
-        fn size(&self) -> usize {
-            (self.end as usize) - (self.cursor as usize)
-        }
-
         #[inline]
         fn take_n(&mut self, ct: usize) -> Result<&'de mut [u8]> {
             let remain = (self.end as usize) - (self.cursor as usize);
             let buff = if remain < ct {
                 return Err(Error::DeserializeUnexpectedEnd);
             } else {
+                // SAFETY: `self.cursor` is valid for `ct` elements and won't be incremented
+                // past `self.end` as we have checked above.
                 unsafe {
                     let sli = core::slice::from_raw_parts_mut(self.cursor, ct);
                     self.cursor = self.cursor.add(ct);
@@ -235,8 +261,24 @@ pub mod io {
             Ok(buff)
         }
 
+        #[inline]
+        fn take_n_temp(&mut self, ct: usize) -> Result<&mut [u8]> {
+            let remain = (self.end as usize) - (self.cursor as usize);
+            let buff = if remain < ct {
+                return Err(Error::DeserializeUnexpectedEnd);
+            } else {
+                unsafe {
+                    let sli = core::slice::from_raw_parts_mut(self.cursor, ct);
+                    sli
+                }
+            };
+
+            Ok(buff)
+        }
+
         fn complete(self) -> Result<&'de mut [u8]> {
             let remain = (self.end as usize) - (self.cursor as usize);
+            // SAFETY: `self.cursor` is valid for `remain` elements
             unsafe { Ok(core::slice::from_raw_parts_mut(self.cursor, remain)) }
         }
     }
@@ -248,7 +290,7 @@ pub mod io {
         use super::SlidingBuffer;
         use crate::{Error, Result};
 
-        /// Wrapper over a [`embedded_io`](crate::eio::embedded_io)::[`Read`](crate::eio::Read) and a sliding buffer to implement the [Flavor] trait
+        /// Wrapper over a [`embedded_io`](crate::eio::embedded_io)::[`Read`](crate::eio::Read) and a sliding buffer to implement the [`Flavor`] trait
         pub struct EIOReader<'de, T>
         where
             T: crate::eio::Read,
@@ -261,7 +303,10 @@ pub mod io {
         where
             T: crate::eio::Read,
         {
-            pub(crate) fn new(reader: T, buff: &'de mut [u8]) -> Self {
+            /// Create a new [`EIOReader`] from a reader and a buffer.
+            ///
+            /// `buff` must have enough space to hold all data read during the deserialisation.
+            pub fn new(reader: T, buff: &'de mut [u8]) -> Self {
                 Self {
                     reader,
                     buff: SlidingBuffer::new(buff),
@@ -280,14 +325,14 @@ pub mod io {
             fn pop(&mut self) -> Result<u8> {
                 let mut val = [0; 1];
                 self.reader
-                    .read(&mut val)
+                    .read_exact(&mut val)
                     .map_err(|_| Error::DeserializeUnexpectedEnd)?;
                 Ok(val[0])
             }
 
             #[inline]
             fn size_hint(&self) -> Option<usize> {
-                Some(self.buff.size())
+                None
             }
 
             #[inline]
@@ -299,15 +344,52 @@ pub mod io {
                 Ok(buff)
             }
 
+            #[inline]
+            fn try_take_n_temp<'a>(&'a mut self, ct: usize) -> Result<&'a [u8]>
+            where
+                'de: 'a,
+            {
+                let buff = self.buff.take_n_temp(ct)?;
+                self.reader
+                    .read_exact(buff)
+                    .map_err(|_| Error::DeserializeUnexpectedEnd)?;
+                Ok(buff)
+            }
+
             /// Return the remaining (unused) bytes in the Deserializer
             fn finalize(self) -> Result<(T, &'de mut [u8])> {
                 let buf = self.buff.complete()?;
                 Ok((self.reader, buf))
             }
         }
+
+        #[cfg(test)]
+        mod tests {
+            use super::*;
+
+            #[test]
+            fn test_pop() {
+                let mut reader = EIOReader::new(&[0xAA, 0xBB, 0xCC][..], &mut []);
+
+                assert_eq!(reader.pop(), Ok(0xAA));
+                assert_eq!(reader.pop(), Ok(0xBB));
+                assert_eq!(reader.pop(), Ok(0xCC));
+                assert_eq!(reader.pop(), Err(Error::DeserializeUnexpectedEnd));
+            }
+
+            #[test]
+            fn test_try_take_n() {
+                let mut buf = [0; 8];
+                let mut reader = EIOReader::new(&[0xAA, 0xBB, 0xCC, 0xDD, 0xEE][..], &mut buf);
+
+                assert_eq!(reader.try_take_n(2), Ok(&[0xAA, 0xBB][..]));
+                assert_eq!(reader.try_take_n(2), Ok(&[0xCC, 0xDD][..]));
+                assert_eq!(reader.try_take_n(2), Err(Error::DeserializeUnexpectedEnd));
+            }
+        }
     }
 
-    /// Support for [std::io] traits
+    /// Support for [`std::io`] traits
     #[allow(clippy::module_inception)]
     #[cfg(feature = "use-std")]
     pub mod io {
@@ -315,7 +397,7 @@ pub mod io {
         use super::SlidingBuffer;
         use crate::{Error, Result};
 
-        /// Wrapper over a [std::io::Read] and a sliding buffer to implement the [Flavor] trait
+        /// Wrapper over a [`std::io::Read`] and a sliding buffer to implement the [Flavor] trait
         pub struct IOReader<'de, T>
         where
             T: std::io::Read,
@@ -328,7 +410,10 @@ pub mod io {
         where
             T: std::io::Read,
         {
-            pub(crate) fn new(reader: T, buff: &'de mut [u8]) -> Self {
+            /// Create a new [`IOReader`] from a reader and a buffer.
+            ///
+            /// `buff` must have enough space to hold all data read during the deserialisation.
+            pub fn new(reader: T, buff: &'de mut [u8]) -> Self {
                 Self {
                     reader,
                     buff: SlidingBuffer::new(buff),
@@ -347,14 +432,14 @@ pub mod io {
             fn pop(&mut self) -> Result<u8> {
                 let mut val = [0; 1];
                 self.reader
-                    .read(&mut val)
+                    .read_exact(&mut val)
                     .map_err(|_| Error::DeserializeUnexpectedEnd)?;
                 Ok(val[0])
             }
 
             #[inline]
             fn size_hint(&self) -> Option<usize> {
-                Some(self.buff.size())
+                None
             }
 
             #[inline]
@@ -366,12 +451,49 @@ pub mod io {
                 Ok(buff)
             }
 
+            #[inline]
+            fn try_take_n_temp<'a>(&'a mut self, ct: usize) -> Result<&'a [u8]>
+            where
+                'de: 'a,
+            {
+                let buff = self.buff.take_n_temp(ct)?;
+                self.reader
+                    .read_exact(buff)
+                    .map_err(|_| Error::DeserializeUnexpectedEnd)?;
+                Ok(buff)
+            }
+
             /// Return the remaining (unused) bytes in the Deserializer
             fn finalize(self) -> Result<(T, &'de mut [u8])> {
                 let buf = self.buff.complete()?;
                 Ok((self.reader, buf))
             }
         }
+
+        #[cfg(test)]
+        mod tests {
+            use super::*;
+
+            #[test]
+            fn test_pop() {
+                let mut reader = IOReader::new(&[0xAA, 0xBB, 0xCC][..], &mut []);
+
+                assert_eq!(reader.pop(), Ok(0xAA));
+                assert_eq!(reader.pop(), Ok(0xBB));
+                assert_eq!(reader.pop(), Ok(0xCC));
+                assert_eq!(reader.pop(), Err(Error::DeserializeUnexpectedEnd));
+            }
+
+            #[test]
+            fn test_try_take_n() {
+                let mut buf = [0; 8];
+                let mut reader = IOReader::new(&[0xAA, 0xBB, 0xCC, 0xDD, 0xEE][..], &mut buf);
+
+                assert_eq!(reader.try_take_n(2), Ok(&[0xAA, 0xBB][..]));
+                assert_eq!(reader.try_take_n(2), Ok(&[0xCC, 0xDD][..]));
+                assert_eq!(reader.try_take_n(2), Err(Error::DeserializeUnexpectedEnd));
+            }
+        }
     }
 }
 
@@ -380,11 +502,11 @@ pub mod io {
 ////////////////////////////////////////
 
 /// This Cyclic Redundancy Check flavor applies [the CRC crate's `Algorithm`](https://docs.rs/crc/latest/crc/struct.Algorithm.html) struct on
-/// the serialized data. The flavor will check the CRC assuming that it has been appended to the bytes.
+/// the serialized data.
 ///
+/// The flavor will check the CRC assuming that it has been appended to the bytes.
 /// CRCs are used for error detection when reading data back.
-///
-/// The `crc` feature requires enabling to use this module.
+/// Requires the `crc` feature.
 ///
 /// More on CRCs: <https://en.wikipedia.org/wiki/Cyclic_redundancy_check>.
 #[cfg(feature = "use-crc")]
@@ -402,7 +524,6 @@ pub mod crc {
     use crate::Deserializer;
     use crate::Error;
     use crate::Result;
-    use paste::paste;
 
     /// Manages CRC modifications as a flavor.
     pub struct CrcModifier<'de, B, W>
@@ -426,94 +547,96 @@ pub mod crc {
     }
 
     macro_rules! impl_flavor {
-        ($( $int:ty ),*) => {
-            $(
-                paste! {
-                    impl<'de, B> Flavor<'de> for CrcModifier<'de, B, $int>
-                    where
-                        B: Flavor<'de>,
-                    {
-                        type Remainder = B::Remainder;
+        ($int:ty, $from_bytes:ident, $take_from_bytes:ident) => {
+            impl<'de, B> Flavor<'de> for CrcModifier<'de, B, $int>
+            where
+                B: Flavor<'de>,
+            {
+                type Remainder = B::Remainder;
+                type Source = B::Source;
 
-                        type Source = B::Source;
-
-                        #[inline]
-                        fn pop(&mut self) -> Result<u8> {
-                            match self.flav.pop() {
-                                Ok(byte) => {
-                                    self.digest.update(&[byte]);
-                                    Ok(byte)
-                                }
-                                e @ Err(_) => e,
-                            }
+                #[inline]
+                fn pop(&mut self) -> Result<u8> {
+                    match self.flav.pop() {
+                        Ok(byte) => {
+                            self.digest.update(&[byte]);
+                            Ok(byte)
                         }
-
-                        #[inline]
-                        fn size_hint(&self) -> Option<usize> {
-                            self.flav.size_hint()
-                        }
-
-                        #[inline]
-                        fn try_take_n(&mut self, ct: usize) -> Result<&'de [u8]> {
-                            match self.flav.try_take_n(ct) {
-                                Ok(bytes) => {
-                                    self.digest.update(bytes);
-                                    Ok(bytes)
-                                }
-                                e @ Err(_) => e,
-                            }
-                        }
-
-                        fn finalize(mut self) -> Result<Self::Remainder> {
-                            match self.flav.try_take_n(core::mem::size_of::<$int>()) {
-                                Ok(prev_crc_bytes) => match self.flav.finalize() {
-                                    Ok(remainder) => {
-                                        let crc = self.digest.finalize();
-                                        let le_bytes = prev_crc_bytes
-                                            .try_into()
-                                            .map_err(|_| Error::DeserializeBadEncoding)?;
-                                        let prev_crc = <$int>::from_le_bytes(le_bytes);
-                                        if crc == prev_crc {
-                                            Ok(remainder)
-                                        } else {
-                                            Err(Error::DeserializeBadCrc)
-                                        }
-                                    }
-                                    e @ Err(_) => e,
-                                },
-                                Err(e) => Err(e),
-                            }
-                        }
-                    }
-
-                    /// Deserialize a message of type `T` from a byte slice with a Crc. The unused portion (if any)
-                    /// of the byte slice is not returned.
-                    pub fn [<from_bytes_ $int>]<'a, T>(s: &'a [u8], digest: Digest<'a, $int>) -> Result<T>
-                    where
-                        T: Deserialize<'a>,
-                    {
-                        let flav = CrcModifier::new(Slice::new(s), digest);
-                        let mut deserializer = Deserializer::from_flavor(flav);
-                        let r = T::deserialize(&mut deserializer)?;
-                        let _ = deserializer.finalize()?;
-                        Ok(r)
-                    }
-
-                    /// Deserialize a message of type `T` from a byte slice with a Crc. The unused portion (if any)
-                    /// of the byte slice is returned for further usage
-                    pub fn [<take_from_bytes_ $int>]<'a, T>(s: &'a [u8], digest: Digest<'a, $int>) -> Result<(T, &'a [u8])>
-                    where
-                        T: Deserialize<'a>,
-                    {
-                        let flav = CrcModifier::new(Slice::new(s), digest);
-                        let mut deserializer = Deserializer::from_flavor(flav);
-                        let t = T::deserialize(&mut deserializer)?;
-                        Ok((t, deserializer.finalize()?))
+                        e @ Err(_) => e,
                     }
                 }
-            )*
+
+                #[inline]
+                fn size_hint(&self) -> Option<usize> {
+                    self.flav.size_hint()
+                }
+
+                #[inline]
+                fn try_take_n(&mut self, ct: usize) -> Result<&'de [u8]> {
+                    match self.flav.try_take_n(ct) {
+                        Ok(bytes) => {
+                            self.digest.update(bytes);
+                            Ok(bytes)
+                        }
+                        e @ Err(_) => e,
+                    }
+                }
+
+                fn finalize(mut self) -> Result<Self::Remainder> {
+                    match self.flav.try_take_n(core::mem::size_of::<$int>()) {
+                        Ok(prev_crc_bytes) => match self.flav.finalize() {
+                            Ok(remainder) => {
+                                let crc = self.digest.finalize();
+                                let le_bytes = prev_crc_bytes
+                                    .try_into()
+                                    .map_err(|_| Error::DeserializeBadEncoding)?;
+                                let prev_crc = <$int>::from_le_bytes(le_bytes);
+                                if crc == prev_crc {
+                                    Ok(remainder)
+                                } else {
+                                    Err(Error::DeserializeBadCrc)
+                                }
+                            }
+                            e @ Err(_) => e,
+                        },
+                        Err(e) => Err(e),
+                    }
+                }
+            }
+
+            /// Deserialize a message of type `T` from a byte slice with a Crc. The unused portion (if any)
+            /// of the byte slice is not returned.
+            pub fn $from_bytes<'a, T>(s: &'a [u8], digest: Digest<'a, $int>) -> Result<T>
+            where
+                T: Deserialize<'a>,
+            {
+                let flav = CrcModifier::new(Slice::new(s), digest);
+                let mut deserializer = Deserializer::from_flavor(flav);
+                let r = T::deserialize(&mut deserializer)?;
+                let _ = deserializer.finalize()?;
+                Ok(r)
+            }
+
+            /// Deserialize a message of type `T` from a byte slice with a Crc. The unused portion (if any)
+            /// of the byte slice is returned for further usage
+            pub fn $take_from_bytes<'a, T>(
+                s: &'a [u8],
+                digest: Digest<'a, $int>,
+            ) -> Result<(T, &'a [u8])>
+            where
+                T: Deserialize<'a>,
+            {
+                let flav = CrcModifier::new(Slice::new(s), digest);
+                let mut deserializer = Deserializer::from_flavor(flav);
+                let t = T::deserialize(&mut deserializer)?;
+                Ok((t, deserializer.finalize()?))
+            }
         };
     }
 
-    impl_flavor![u8, u16, u32, u64, u128];
+    impl_flavor!(u8, from_bytes_u8, take_from_bytes_u8);
+    impl_flavor!(u16, from_bytes_u16, take_from_bytes_u16);
+    impl_flavor!(u32, from_bytes_u32, take_from_bytes_u32);
+    impl_flavor!(u64, from_bytes_u64, take_from_bytes_u64);
+    impl_flavor!(u128, from_bytes_u128, take_from_bytes_u128);
 }
